import os
import os.path
import random
import string
import subprocess
import charms.leadership
from base64 import b64decode
from charms.reactive import clear_flag, endpoint_from_flag, set_flag, when, when_any, when_not, hook
from charmhelpers.core import hookenv, unitdata
from charmhelpers.core.host import mkdir, service, write_file
from charmhelpers.core.templating import render
from charmhelpers.contrib.charmsupport import nrpe


@hook('upgrade-charm')
def upgrade_charm():
    clear_flag('mailman.database.ran_migrations')
    clear_flag('mailman.apache.configured')
    clear_flag('mailman.postfix.configured')
    write_config()


# Pull all the keys out of the kv() key/value store and shove them into a dict
# that we can render into our config, then restart mailman3-web.
def write_config():
    db = unitdata.kv()
    config = hookenv.config()
    config.update(db.getrange('django_'))
    config.update(db.getrange('admin_'))
    config.update(db.getrange('db_'))
    render(source="mailman-web.tmpl",
           target="/etc/mailman3/mailman-web.py",
           owner="www-data",
           group="list",
           perms=0o640,
           context={'config': config})
    render(source="uwsgi.tmpl",
           target="/etc/mailman3/uwsgi.ini",
           owner="root",
           group="root",
           perms=0o644,
           context={'config': config})
    service('restart', 'mailman3-web')
    clear_flag('mailman.database.configured_site')


# Generate the Django secret key and HyperKitty's API key and put them in kv().
@when('apt.installed.mailman3-web')
@when('leadership.is_leader')
@when_not('leadership.set.django_secret_key')
def generate_django_secret():
    chars = string.ascii_letters + string.digits
    django_secret_key = ''.join((random.choice(chars)) for x in range(40))
    charms.leadership.leader_set(django_secret_key=django_secret_key)


@when('apt.installed.mailman3-web')
@when('leadership.is_leader')
@when_not('leadership.set.django_archiver_key')
def generate_hyperkitty_archiver_secret():
    chars = string.ascii_letters + string.digits
    hyperkitty_archiver_key = ''.join((random.choice(chars)) for x in range(40))
    charms.leadership.leader_set(django_archiver_key=hyperkitty_archiver_key)


@when_any('leadership.changed.django_secret_key', 'leadership.changed.django_archiver_key')
def store_secrets():
    db = unitdata.kv()
    db.set('django_secret_key', hookenv.leader_get('django_secret_key'))
    db.set('django_archiver_key', hookenv.leader_get('django_archiver_key'))
    db.set('archiver_ip', hookenv.unit_get('private-address'))
    clear_flag('mailman.apache.configured')


# We're going to use Apache as a uWSGI proxy because for reasons I don't fully
# understand, uWSGI itself is dog slow as a HTTP server but proxied via Apache
# it's reasonably quick.  So write/enable all the configs we need then enable
# Apache as a service and give it a restart so it picks up the new configs.
@when('apt.installed.apache2')
@when_not('mailman.apache.configured')
def configure_apache2():
    hookenv.status_set('maintenance', 'updating apache2 configuration')
    render(source="mailman-apache.tmpl",
           target="/etc/apache2/conf-available/mailman.conf",
           context={})
    # mailman3-web started writing this file sometime during the bionic release
    # cycle so check for it and if it's there, nuke it and install our known
    # good one.
    if os.path.isfile("/etc/apache2/conf-enabled/mailman.conf"):
        os.unlink("/etc/apache2/conf-enabled/mailman.conf")
    subprocess.check_output(['a2enconf', 'mailman'])
    # proxy-uwsgi is a dependency of mailman3-web, we just need to enable it.
    subprocess.check_output(['a2enmod', 'proxy', 'proxy_uwsgi'])

    service('enable', 'apache2')
    service('reload', 'apache2')
    # Open the port while we're at it, we're assuming something else will
    # do SSL termination in front of this.
    hookenv.open_port(80)
    set_flag('mailman.apache.configured')
    hookenv.status_set('active', 'ready')


# Ask postgresql to create our DB.
@when_not('mailman.database.created')
@when('db.connected')
def request_db(pgsql):
    pgsql.set_database('mailman3web')
    set_flag('mailman.database.created')


@when('apt.installed.mailman3-web')
@when('leadership.is_leader')
@when('mailman.database.configured')
@when_not('mailman.database.ran_migrations')
def perform_db_migrations():
    hookenv.status_set('maintenance', 'running Django data migrations to synchronize DB state')
    subprocess.check_output(['sudo', '-u', 'www-data', '/usr/share/mailman3-web/manage.py', 'migrate'])
    set_flag('mailman.database.ran_migrations')
    hookenv.status_set('active', 'ready')


@when('apt.installed.mailman3-web')
@when('leadership.is_leader')
@when('mailman.database.configured')
@when_not('mailman.database.configured_site')
def configure_django_site():
    hookenv.status_set('maintenance', 'configuring Django site')
    config = hookenv.config()
    subprocess.check_output(['sudo', '-u', 'www-data', '/usr/share/mailman3-web/manage.py', 'set_default_site', '--domain', config['domain'], '--name', config['site_name']])
    set_flag('mailman.database.configured_site')
    hookenv.status_set('active', 'ready')


# And fetch DB connection details, shove them in kv().
@when('db.master.available')
@when_not('mailman.database.configured')
def configure_database(pgsql):
    db_relation_config = pgsql.master
    db = unitdata.kv()
    db.set('db_name', db_relation_config.dbname)
    db.set('db_host', db_relation_config.host)
    db.set('db_user', db_relation_config.user)
    db.set('db_password', db_relation_config.password)
    write_config()
    set_flag('mailman.database.configured')


@when('apt.installed.postfix')
@when_not('mailman.postfix.configured')
def configure_postfix():
    hookenv.status_set('maintenance', 'configuring postfix for mail delivery')
    subprocess.check_output(['postconf', 'smtp_address_preference=ipv4',
                             'default_transport=smtp', 'relay_transport=smtp',
                             'smtpd_relay_restrictions=permit_mynetworks defer_unauth_destination'])
    subprocess.check_output(['postfix', 'reload'])
    set_flag('mailman.postfix.configured')
    hookenv.status_set('active', 'ready')


# And once the DB is actually available, write the config and restart.
@when('db.master.available')
@when('config.changed')
def write_updated_config():
    hookenv.status_set('maintenance', 'updating configuration')
    write_config()
    hookenv.status_set('active', 'ready')


# We're providing a website relation to whatever does SSL termination in front
# of us.
@when('mailman3website.available')
def configure_mailman3website(mailman3website):
    mailman3website.configure(port=80, hostname=hookenv.config('domain'))


# This is a tiny bit odd, the relation.joined state is only cleared when the
# last related unit leaves the relation, but when it's initially set, i.e. on
# the relation actually being joined by the first related unit, it doesn't have
# any data.  So we guard the reaction with a relation.configured flag, it runs
# once on join with no data, the guard flag gets set, then when the relation
# changes, the interface sets relation.connected and we clear the guard flag
# so this can re-run.  In this case we're providing the relation, not consuming
# so it's slightly less relevant, but we still need to follow the same pattern.
@when('endpoint.mailman3archiver.joined')
@when_not('mailman3archiver.relation.configured')
def configure_mailman3archiver():
    db = unitdata.kv()
    ip = db.get('archiver_ip')
    port = 80
    user = ''
    password = db.get('django_archiver_key')
    mailman3archiver = endpoint_from_flag('endpoint.mailman3archiver.joined')
    mailman3archiver.publish_info(ip, port, user, password)
    set_flag('mailman3archiver.relation.configured')


# When the last unit leaves the relation, clear the guard flag so when/if one
# rejoins, the relation publish will run.
@when_not('endpoint.mailman3archiver.joined')
def clear_archiver_relation():
    clear_flag('mailman3archiver.relation.configured')


@when_any('endpoint.mailman3archiver.connected')
def changed_archiver_data():
    clear_flag('mailman3archiver.relation.configured')


# This works broadly similarly to the mailman3archiver relation above, except
# in this case we're consuming the relation rather than providing.
@when('endpoint.mailman3.joined')
@when_not('mailman3.relation.configured')
def configure_mailman3():
    db = unitdata.kv()
    mailman3_relations = endpoint_from_flag('endpoint.mailman3.joined')
    mailman3_data = mailman3_relations.mailman3()[0]
    db.set('admin_ip', mailman3_data['ip'])
    db.set('admin_port', mailman3_data['port'])
    db.set('admin_user', mailman3_data['user'])
    db.set('admin_password', mailman3_data['password'])
    write_config()
    set_flag('mailman3.relation.configured')


@when_not('endpoint.mailman3.joined')
def clear_mailman3_relation():
    clear_flag('mailman3.relation.configured')


@when_any('endpoint.mailman3.connected')
def changed_mailman_data():
    if hookenv.hook_name() == 'update-status':
        return
    clear_flag('mailman3.relation.configured')


@when('nrpe-external-master.available')
def setup_nagios(nagios):
    if hookenv.hook_name() == 'update-status':
        return

    hookenv.status_set('maintenance', 'setting up NRPE checks')
    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    config = hookenv.config()

    check_base = '/usr/lib/nagios/plugins/'
    process_check = check_base + 'check_procs'

    uwsgi_processes = config['uwsgi_processes'] + 1  # plus the master process
    wsgi_process_check = '{} -c {}:{} -C uwsgi'.format(process_check, uwsgi_processes, uwsgi_processes)
    nrpe_setup.add_check('wsgi_process', 'Check for Mailman3 Django WSGI process', wsgi_process_check)

    if config['check-lists-command'] != '':
        # Install check_list_settings as a django management command, and its config file in /etc/nagios
        management_path = '/usr/lib/python2.7/dist-packages/django_mailman3/management'
        command_path = os.path.join(management_path, 'commands')
        for path in [management_path, command_path]:
            mkdir(path, perms=0o755)
            write_file(os.path.join(path, '__init__.py'), "default_app_config = 'django_mailman3.apps.DjangoMailman3Config'", owner='root', group='root', perms=0o644)
        command_file = os.path.join(command_path, 'check_list_settings.py')
        render(config_template=b64decode(config['check-lists-command']).decode('utf-8'),
               source=None,
               target=command_file,
               owner='root',
               group='root',
               perms=0o644,
               context={'config': config})
        write_file('/etc/nagios/check-lists.conf',
                   b64decode(config['check-lists-conf']),
                   owner='nagios',
                   group='nagios',
                   perms=0o644)
        # Cron it
        cronpath = '/etc/cron.d/check-lists'
        checkpath = '/var/lib/nagios/list-config-sanity.txt'
        croncmd = 'python /usr/share/mailman3-web/manage.py check_list_settings --verbose'
        cronfile = '*/15 * * * * www-data {croncmd} > {checkpath}.new; cp {checkpath}.new {checkpath}\n'.format(croncmd=croncmd, checkpath=checkpath)
        write_file(cronpath, cronfile, owner='root', group='root', perms=0o644)
        write_file(checkpath, '', owner='www-data', group='www-data', perms=0o644)
        write_file('{}.new'.format(checkpath), '', owner='www-data', group='www-data', perms=0o644)
        # TODO: use charmhelpers' check_status_file.py, once it supports file max age as a parameter
        nrpe.copy_nrpe_checks(os.path.join(os.getenv('CHARM_DIR'), 'files'))
        # And add the check
        nrpe_setup.add_check(
            shortname="list_settings",
            description="Mailing list config sanity check",
            check_cmd='/usr/local/lib/nagios/plugins/check_cronned_output.py %s' % checkpath
        )

    nrpe_setup.write()
    hookenv.status_set('active', 'ready')
